#include <math.h>
#include "image.h"

void InitImage(unsigned short *dst, int width, int height, unsigned short val){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            *(dst+i+width*j) = val;
        }
    }
}


//コピーもとのメモリから格納先メモリに比較明しながら保存します。
void SLImage(unsigned short *dst, unsigned short *src, int width, int height, int num){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(dst+i+width*j) < *(src+i+width*j))
                *(dst+i+width*j) = *(src+i+width*j);
        }
    }
}


//コピーもとのメモリから格納先メモリに比較暗しながら保存します。
void SDImage(unsigned short *dst, unsigned short *src, int width, int height, int num){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(dst+i+width*j) > *(src+i+width*j))
                *(dst+i+width*j) = *(src+i+width*j);
        }
    }
}


void AvrImage(unsigned short *dst, unsigned short *src, int width, int height, int num){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            *(dst+i+width*j) = (*(dst+i+width*j) * num + *(src+i+width*j))/(num+1);
        }
    }
}


void WBImageRG(unsigned short *src, int width, int height){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(i%2==0 && j%2==0){
                *(src+i+width*j) = *(src+i+width*j)/2;
            }else if(i%2==1 && j%2==1){
                ;
            }else{
                *(src+i+width*j) = *(src+i+width*j)/2;
            }
        }
    }
}


void SubImage(unsigned short *dst, unsigned short *src, int width, int height, int num){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(dst+i+width*j) > *(src+i+width*j))
                *(dst+i+width*j) -= *(src+i+width*j);
            else
                *(dst+i+width*j) = 0;
        }
    }
}


void DifImage(unsigned short *dst, unsigned short *src1, unsigned short *src2, int width, int height, int num){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(src1+i+width*j) > *(src2+i+width*j))
                *(dst+i+width*j) = *(src1+i+width*j) - *(src2+i+width*j);
            else if(num == 0)
                *(dst+i+width*j) = 0;
            else
                *(dst+i+width*j) = *(src2+i+width*j) - *(src1+i+width*j);
        }
    }
}


void AbsImage(unsigned short *dst, unsigned short *src, int width, int height, int num){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(dst+i+width*j) >= *(src+i+width*j))
                *(dst+i+width*j) -= *(src+i+width*j);
            else
                *(dst+i+width*j) = *(src+i+width*j) - *(dst+i+width*j);
        }
    }
}


void CopyImage(unsigned short *dst, unsigned short *src, int width, int height, int num){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            *(dst+i+width*j) = *(src+i+width*j);
        }
    }
}


float BinzImage(unsigned short *dst, unsigned short threshold, int width, int height){
    int count = 0;
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(dst+i+width*j) > threshold){
                *(dst+i+width*j) = 0xffff;
                count++;
            }else{
                *(dst+i+width*j) = 0;
            }
        }
    }
    return (float)count/(float)(width * height);
}

float CountImage(unsigned short *dst, unsigned short threshold, int width, int height){
    int count = 0;
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(dst+i+width*j) > threshold){
                count++;
            }
        }
    }
    return (float)count/(float)(width * height);
}

void SDeviImage(unsigned short *dst, unsigned short threshold, int width, int height, double *ax, double *ay, double *sx, double *sy){
    int count = 0;
    int sumx = 0;
    int sumy = 0;
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(dst+i+width*j) > threshold){
                count++;
                sumx += i;
                sumy += j;
            }
        }
    }
    *ax = (double)sumx / (double)count;
    *ay = (double)sumy / (double)count;
    double qsumx = 0;
    double qsumy = 0;
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(dst+i+width*j) > threshold){
                qsumx += ((double)i - *ax)*((double)i - *ax);
                qsumy += ((double)j - *ay)*((double)j - *ay);
            }
        }
    }
    *sx = sqrt(qsumx/(double)count);
    *sy = sqrt(qsumy/(double)count);
}

void MaxMinImage(unsigned short *dst, unsigned short threshold, int width, int height, int *minx, int *miny, int *maxx, int *maxy){
    *minx = width;
    *miny = height;
    *maxx = 0;
    *maxy = 0;
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(*(dst+i+width*j) > threshold){
                *minx = (i < *minx) ? i : *minx;
                *miny = (j < *miny) ? j : *miny;
                *maxx = (i > *maxx) ? i : *maxx;
                *maxy = (j > *maxy) ? j : *maxy;
            }
        }
    }
}

void NRImage(unsigned short *dst, unsigned short threshold, int width, int height){
#pragma omp parallel for
    for(int j=0; j<height; j++){
#pragma omp parallel for
        for(int i=0; i<width; i++){
            if(i>0 && i<width-1 && j>0 && j<height-1){
                if(*(dst+i+width*j) > threshold){
                    if(*(dst+i-1+width*j) < threshold && *(dst+i+1+width*j) < threshold
                        && *(dst+i+width*(j-1)) < threshold && *(dst+i+width*(j+1)) < threshold){
                        *(dst+i+width*j) = *(dst+i-1+width*j);
                    }
                }
            }
        }
    }
}
