//#define SC132GS
//#define IMX219
/*====================================================================*
 * sensorlaunch.c
 *
 * Copyright(c) 2020-2021 Soho Enterprise Ltd. (https://soho-enterprise.com/)
 *====================================================================*/
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <opencv2/opencv.hpp>
#include <opencv2/highgui.hpp>
#include <math.h>
#include "zoom.h"
#include "fits.h"
#include "ser.h"
#include "image.h"

static const char __PROGRAM_VERSION__[] = "@(#)sensorlaunch.c Ver. 1.0036 (2022/03/01 16:43:03)";

int     OpenSensor(char *fname);
int     CloseSensor(int status);
int     ReadRAW(unsigned char *RawBuf);
int     GetWidth();
int     GetHeight();
int     GetBufSiz();
int     SetReg(int reg16, int len, int data);
//
char    *MakeFileName(int width, int height);
int     WriteRAW(char *fname, unsigned char *raw, int size);
double  GetTime();
//
void    camera_i2c(int device_no);
int     magic();
//
int     check_rawbit(char *fname);

int     frame_count, oc;
double  st, et;

//#define TIMEOUT   (3000)
//#define TIMEOUT   (60000)
#define TIMEOUT   (300000)

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
void
SetInit()
{
#ifdef SC132GS
    SetReg(0x3e00, 3, 1000);         // Exposure
    SetReg(0x3e08, 2, 256);          // AnaGain
    //SetReg(0x3221, 1, 0x66);       // Flip&Mirror
#endif
#ifdef IMX219
    SetReg(0x015A, 2, expval); // Exposure
    SetReg(0x0157, 1, anaval); // AnaGain
    SetReg(0x0158, 2, digval); // DigGain
    SetReg(0x0172, 1, flip  ); // Flip&Mirror
#endif
}

//コピーもとのメモリから格納先メモリに保存します。
//packされた10bit,12bitは16bitに展開してします。
void CnvertImage(unsigned char *dst, unsigned char *src, int srcSize, int bpp, int raw8){
	int j = 0;
    if(raw8){
	switch(bpp){
		case 8:
			for(int i = 0; i < srcSize; i++,j++){
				dst[i] = src[i];
			}
			break;
		case 16:
			for(int i = 0; i < srcSize; i+=2){
				dst[j++] = src[i+1];
			}
			break;
		case 10:
			for(int i = 0; i < srcSize; i+=5){
				dst[j++] = src[i];
				dst[j++] = src[i+1];
				dst[j++] = src[i+2];
				dst[j++] = src[i+3];
			}
			break;
		case 12:
			for(int i = 0; i < srcSize; i+=3){
				dst[j++] = src[i+0];
				dst[j++] = src[i+1];
			}
			break;
		default:
			break;
        }
    }else{
	switch(bpp){
		case 8:
		case 16:
			for(int i = 0; i < srcSize; i++){
				dst[i] = src[i];
			}
			break;
		case 10:
			for(int i = 0; i < srcSize; i+=5){
				dst[j++]  = src[i]    *  4 + (src[i+4] )     % 4;
				dst[j++]  = src[i]    / 64;
				dst[j++]  = src[i+1]  * 4 + (src[i+4] >> 2)  % 4;
				dst[j++]  = src[i+1]  / 64;
				dst[j++]  = src[i+2]  * 4 + (src[i+4] >> 4 ) % 4;
				dst[j++]  = src[i+2]  / 64;
				dst[j++]  = src[i+3]  * 4 + (src[i+4] >> 6 ) % 4;
				dst[j++]  = src[i+3]  / 64;
			}
			break;
		case 12:
			for(int i = 0; i < srcSize; i+=3){
				dst[j++] = src[i]   * 16 + (src[i+2] )     % 16;
				dst[j++] = src[i]   / 16;
				dst[j++] = src[i+1] * 16 + (src[i+2] >> 4) % 16;
				dst[j++] = src[i+1] / 16;
			}
			break;
		default:
			break;
	}
    }
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
void bmask(cv::Mat msk, unsigned short *dst){
#pragma omp parallel for
    for(int j=0; j<msk.rows; j++){
#pragma omp parallel for
        for(int i=0; i<msk.cols; i++ ){
            if(msk.data[ j * msk.step + i * msk.elemSize() ] > 0)   // マスク作成の都合上黒以外を消します
                *(dst+i+msk.cols*j) = 0;
        }
    }
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
#define WAIT_MODE 0
#define SERS_MODE 1
#define SERD_MODE 2
#define SERE_MODE 3
#define QUIT_MODE 4
#define OFF	0
#define ON	1

int
main(int ac, char *av[])
{
    int     s_status;
    int     keycode;
    int     width, height, bufsiz;
    int     xwidth;
    char    *fname;
    int     div;
    int     rawbit;
    char    wfname[BUFSIZ];
    char    tfname[BUFSIZ];
    char    pfname[BUFSIZ];
    bool    raw8_flag = false;
    int     timeout_counter;
    int     mode = WAIT_MODE;
    int     colorID = SER_COLORID_MONO;
    int     diff = ON;            // 差分検出フラグ
    int     diffcount = 0;        // モード内での差分検知カウント
    int     bthresh = 0x6000;     // ノイズ除去と二値化用の閾値
    cv::Mat readmask;
    cv::Mat mask_r;               // 記録エリア用マスクデータ
    cv::Mat mask_d;               // 判定エリア用マスクデータ
    char str[256];

    printf("%s\n", __PROGRAM_VERSION__+4);

    do {
        printf("exec camera_i2c\n");
        camera_i2c(0);
    } while (magic() < 0);

    if (ac == 1) {
        printf("usage> %s sensor_init_file [-8]\n", av[0]);
        return  0;
    } else if (ac == 2) {
        fname = av[1];
    } else if (ac == 3) {
        if (strcmp(av[2], "-8") == 0) {
            raw8_flag = true;
            fname = av[1];
        } else if (strcmp(av[1], "-8") == 0) {
            raw8_flag = true;
            fname = av[2];
        } else {
            printf("usage> %s sensor_init_file [-8]\n", av[0]);
            return  0;
        }
    } else {
        printf("usage> %s sensor_init_file [-8]\n", av[0]);
        return  0;
    }

    if ((rawbit = check_rawbit(fname)) < 0) {
        fprintf(stderr, "sensorlaunch: check_rawbit error\n");
        return  1;
    }

    if(strstr(fname, "imx397") == NULL){
        printf("Color");
	colorID = SER_COLORID_BAYER_RGGB;
    }else{
        printf("Monochrome");
    }
        printf(" Imagesensor Detect\n");

    //-------------------------------------------------------------------
    // イメージセンサーをオープンする
    //-------------------------------------------------------------------
    s_status = OpenSensor(fname);
    if (s_status < 0) {
        (void)CloseSensor(s_status);
        return  1;
    }
    SetInit();

    width  = GetWidth();
    height = GetHeight();
    bufsiz = GetBufSiz();
    xwidth = bufsiz / height;

    if (xwidth > 3800 || height > 3000) {
        div = 4;
    } else if (xwidth > 1900 || height > 1000) {
        div = 2;
    } else {
        div = 1;
    }

    int BinSize;
    int RawSize = bufsiz;
    int bpp = RawSize * 8 / width / height;
    int dpp;                            //変換後のビット深度
    int depth = CV_16UC1;

    //変換後のサイズを計算
    if(bpp == 10){
        BinSize = RawSize * 2 * 4 / 5;	// 8,8,8,8,(2,2,2,2)
        dpp = 16;
    }else if(bpp == 12){
        BinSize = RawSize * 2 * 2 / 3;	// 8,8,(4,4)
        dpp = 16;
    }else{
        BinSize = RawSize;
        dpp = 8;
    }

    if (raw8_flag) {
        if(bpp == 10){
            BinSize = RawSize * 4 / 5;	// 8,8,8,8,(2,2,2,2)
        }else if(bpp == 12){
            BinSize = RawSize * 2 / 3;	// 8,8,(4,4)
        }else{
            BinSize = RawSize;
        }
        dpp = 8;
        depth = CV_8UC1;
    }

    printf("-------------------------\n");
    printf("[Esc] / [q] : quit\n");
    printf("[w]         : write raw data\n");
    printf("[s]         : start recording\n");
    printf("[d]         : start diff recording\n");
    printf("[e]         : end recording\n");
    printf("-------------------------\n");

#ifdef OPENCV4
    cv::namedWindow("se_cam", cv::WINDOW_AUTOSIZE);
#else
    cv::namedWindow("se_cam", CV_WINDOW_AUTOSIZE);
#endif

    //-------------------------------------------------------------------
    // RAWデータを格納する領域を確保する
    //-------------------------------------------------------------------
    unsigned char *RawImage = (unsigned char *)malloc(RawSize);
    unsigned char *BinImage = (unsigned char *)malloc(BinSize);     // 未使用？
    unsigned char *BinImage2 = (unsigned char *)malloc(BinSize);
    unsigned char *BinImage3 = (unsigned char *)malloc(BinSize);
    unsigned char *DRawImage = (unsigned char *)malloc(RawSize);    // 未使用？
    unsigned char *DBinImage = (unsigned char *)malloc(BinSize);    // 未使用？
    unsigned char *DiffImage = (unsigned char *)malloc(BinSize);
    unsigned char *StkImage = (unsigned char *)malloc(BinSize);
    if(RawImage == NULL || BinImage == NULL || BinImage2 == NULL || DRawImage == NULL || DBinImage == NULL){
        perror("malloc error");
        exit(EXIT_FAILURE);
    }

    //-------------------------------------------------------------------
    // 記録エリア用マスクを取得する
    //-------------------------------------------------------------------
    readmask = cv::imread("./mask_rec.png");
    if(readmask.data != NULL)
        cv::cvtColor(readmask, mask_r, cv::COLOR_RGB2GRAY);

    //-------------------------------------------------------------------
    // FITSデータの保存準備する
    //-------------------------------------------------------------------
    InitFits(BinSize, width, height, fname);

    //-------------------------------------------------------------------
    // FITSデータの保存準備する
    //-------------------------------------------------------------------
    InitSer(dpp, width, height, fname, colorID);

    //-------------------------------------------------------------------
    // 画像を読み込み,表示する
    //-------------------------------------------------------------------
    timeout_counter = 0;
    frame_count = oc = 0;
    st = GetTime();

    zoom_setMouseCallback((char *)"se_cam", 1/(float)div);

    while (1) {
        if (ReadRAW(RawImage) == 0) {
            if (++timeout_counter > TIMEOUT) {
                fprintf(stderr, "sensorlaunch: time out.\n");
                return  1;
            }
            usleep(500);
            continue;
        }
        frame_count++;
        timeout_counter = 0;

        unsigned char *Temp = BinImage3;
        BinImage3 = BinImage2;
        BinImage2 = Temp;
        CnvertImage(BinImage2, RawImage, RawSize, rawbit, raw8_flag);
        if(mask_r.data != NULL)
            bmask(mask_r, (unsigned short *)BinImage2);

        if(mode == SERD_MODE){
            DifImage((unsigned short *)DiffImage, (unsigned short *)BinImage2, (unsigned short *)BinImage3, width, height, 0);
            NRImage((unsigned short *)DiffImage, bthresh >> (16-bpp), width, height);
            if(mask_d.data != NULL)
                bmask(mask_d, (unsigned short *)DiffImage);
            BinzImage((unsigned short *)DiffImage, bthresh >> (16-bpp), width, height);
            float lrate = CountImage((unsigned short *)DiffImage, bthresh >> (16-bpp), width, height) * 100;
            if(lrate > 0.001){
                double ax, ay, sx, sy;
                SDeviImage((unsigned short *)DiffImage, bthresh >> (16-bpp), width, height, &ax, &ay, &sx, &sy);
                if(sx/(double)width < 0.1 || sy/(double)height < 0.1){
                    printf("%5d:Diff Detect. sx:%04.0lf sy:%04.0lf\n", ++diffcount, sx, sy);
                    diff = ON;
                    if(frame_count != oc +1)
                        system("~/detect_voice.sh");
                    oc = frame_count;
                }
            }

            // 差分表示用
            cv::Mat bayerXBitMat2(height, width, depth, DiffImage);         // Xbit用バッファを用意して取り込み
            cv::Mat bayer8BitMat2 = bayerXBitMat2.clone();                  // 8bit用バッファを用意
            resize(bayer8BitMat2, bayer8BitMat2, cv::Size(), 1/(float)div, 1/(float)div);       // 縮小
            cv::imshow("ImageDiff", bayer8BitMat2);                          // ウィンドウに画像を表示する

            // 差分積算表示用
            if(diff == ON){
                SLImage((unsigned short *)StkImage, (unsigned short *)DiffImage, width, height, 0);
                cv::Mat bayer8BitMat3(height, width, CV_16UC1, StkImage);   // 16bit用バッファを用意して取り込み
                bayer8BitMat3.convertTo(bayer8BitMat3, CV_8UC1, 0.0625*2);  // 8bitに変換
                sprintf(str, "%d differences detected", diffcount);
                cv::putText(bayer8BitMat3, str, cv::Point(20, 50),cv::FONT_HERSHEY_SIMPLEX, 1, CV_RGB(200,200,200), 1);
                resize(bayer8BitMat3, bayer8BitMat3, cv::Size(), 1/(float)div, 1/(float)div);   // 縮小
                cv::imshow("ImageDiffStack", bayer8BitMat3);                 // ウィンドウに画像を表示する
            }
        }

        // メインのRAW表示用
        cv::Mat bayerXBitMat(height, width, depth, BinImage2);              // Xbit用バッファを用意して取り込み
        cv::Mat bayer8BitMat = bayerXBitMat.clone();                        // 8bit用バッファを用意
        if(!raw8_flag){
//            bayer8BitMat.convertTo(bayer8BitMat, CV_8UC1, 0.0625*4);          // 8bitに変換
            bayer8BitMat.convertTo(bayer8BitMat, CV_8UC1, 0.02*4);          // 8bitに変換
        }
        cv::Mat rgb8BitMat(height, width, CV_8UC3);                         // 保存用RGBバッファを用意
        cv::Mat dsp8BitMat(height, width, CV_8UC3);	                        // 表示用RGBバッファを用意
        cv::cvtColor(bayer8BitMat, rgb8BitMat, cv::COLOR_BayerRG2RGB);      // デベイヤー
        resize(rgb8BitMat, dsp8BitMat, cv::Size(), 1/(float)div, 1/(float)div);     // 縮小
        zoom_imshow(rgb8BitMat, dsp8BitMat, 2.0);                           // 拡大表示ウィンドウに画像を表示する
        cv::imshow("se_cam", dsp8BitMat);                                   // ウィンドウに画像を表示する

        // キー入力の処理
        keycode = cv::waitKey(1);
        if (keycode == 27 || keycode == 'q') {
            if(mode == WAIT_MODE)
                break;
            else
                mode = QUIT_MODE;
        } else if (keycode == 'r') {                                        // 再読み込み
            readmask = cv::imread("./mask_rec.png");
            if(readmask.data != NULL)
                cv::cvtColor(readmask, mask_r, cv::COLOR_RGB2GRAY);
        } else if (keycode == 'w') {
            sprintf(tfname, "%s", MakeFileName(width, height));
            sprintf(wfname, "%s.raw", tfname);
            WriteRAW(wfname, (unsigned char *)BinImage2, BinSize);
            printf("write: %s\n", wfname);
            sprintf(wfname, "%s.png", tfname);
            cv::imwrite(wfname, rgb8BitMat);
            printf("write: %s\n", wfname);
            sprintf(wfname, "%s.fts", tfname);
            printf("write: %s\n", wfname);
            WriteFTS(BinImage2, BinSize, bpp, wfname);
        } else if(keycode == 's' || keycode == 'd') {
            if(mode == WAIT_MODE){
                if(keycode == 's'){
                    mode = SERS_MODE;
                }else{
                    mode = SERD_MODE;
                    diffcount = 0;
                    readmask = cv::imread("./mask_det.png");
                    InitImage((unsigned short *)StkImage, width, height, 0);
                    if(readmask.data != NULL)
                    cv::cvtColor(readmask, mask_d, cv::COLOR_RGB2GRAY);
                }
                sprintf(tfname, "%s", MakeFileName(width, height));
                sprintf(wfname, "%s.ser", tfname);
                WriteSERHeader(wfname);
            }
        } else if(keycode == 'e') {
            if(mode == SERS_MODE){
                mode = SERE_MODE;
            }else if(mode == SERD_MODE){
                mode = SERE_MODE;
                diff = OFF;
                cv::destroyWindow("ImageDiff");

                // 差分積算をサムネイルとして保存
                if(diffcount != 0){
                    cv::Mat bayerMat(height, width, CV_16UC1, StkImage);    // 16bit用バッファを用意して取り込み
                    bayerMat.convertTo(bayerMat, CV_8UC1, 0.0625*2);        // 8bitに変換
                    sprintf(str, "%d differences detected", diffcount);
                    cv::putText(bayerMat, str, cv::Point(20, 50),cv::FONT_HERSHEY_SIMPLEX, 1, CV_RGB(200,200,200), 1);
                    sprintf(pfname, "%s_%04ddifs.png", tfname, diffcount); // ファイル名に検知数を入れておく
                    cv::imwrite(pfname, bayerMat);                          // PNGで保存
                    cv::destroyWindow("ImageDiffStack");
                }
            }
        }

        // 順番があるためまとめて処理
        if(mode != WAIT_MODE) {
            if(mode != SERD_MODE || mode == SERD_MODE && diff == ON){
                //フレーム書込み
                WriteSERImage(BinImage2, width * height * dpp / 8, dpp);
                diff = OFF;
            }
            if(mode == SERE_MODE) {
                //終了処理
                CloseSER();
                mode = WAIT_MODE;
            } else if(mode == QUIT_MODE) {
                //終了処理
                CloseSER();
                break;
            }
        }
    }
    et = GetTime();
    cv::destroyWindow("se_cam");

    //-------------------------------------------------------------------
    // イメージセンサーをクローズする
    //-------------------------------------------------------------------
    if (CloseSensor(s_status) < 0) {
        printf(" Error CloseSensor.\n");
        return  1;
    }

    printf("frame_count = %d\n", frame_count);
    printf("time        = %f\n", et-st);
    printf("fps         = %f\n", 1.0/((et-st)/(double)frame_count));

    return  0;
}
